unit CsvUtils2;

interface
uses
  Windows, SysUtils, Classes;

type
  PCsvProgressFunc = ^TCsvProgressFunc;
  TCsvProgressFunc = procedure (readBytes, totalBytes: Integer; var Cancel: Boolean) of Object;

  TCsvCells = class
  private
    list: TList;
    function GetCount: Integer;
    function GetValue(Index: Integer): string;
    procedure SetValue(Index: Integer; const Value: string);
    procedure SetCount(const Value: Integer);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear; virtual;
    procedure Delete(Index: Integer);
    function GetCommaText: string;
    function GetCommaTextEx(Splitter: Char; ColCount: Integer = -1): string;
    function GetTabText: string;
    function Add(Value: string): Integer;
    property Count: Integer read GetCount write SetCount;
    procedure Insert(Index: Integer; Value: string);
    procedure Move(FromI, ToI: Integer);
    property Values[Index: Integer]: string read GetValue write SetValue;
    procedure Assign(Source: TCsvCells);
    procedure TrimLast;
    function AllBlank: Boolean;
  end;

  TCsvSheet = class
  private
    list: TList;
    FUseHeader: Boolean;
    function GetCount: Integer;
    function GetCell(ACol, ARow: Integer): string;
    procedure SetCell(ACol, ARow: Integer; const Value: string);
    procedure SetCount(const Value: Integer);
    function GetColCount: Integer;
    function GetAsText: string;
    procedure SetAsText(const Value: string);
    function GetAsTabText: string;
    procedure SetAsTabText(const Value: string);
    procedure SetAsTextCustom(const Delimiter: Char; const Value: string);
    procedure SetColCount(const Value: Integer);
    function GetRow(ARow: Integer): TCsvCells;
  public
    constructor Create;
    destructor Destroy; override;
    // ǉړ폜
    procedure Clear; virtual;
    procedure InsertRow(Index: Integer); // s}
    procedure InsertCol(Index: Integer); // s}
    procedure DeleteRow(Index: Integer);
    procedure DeleteCol(Index: Integer);
    procedure MoveRow(FromI, ToI: Integer);
    procedure MoveCol(FromI, ToI: Integer);
    procedure TrimBottom; // [s̃g
    procedure TrimRight;  // E[̃g
    // EDIT
    procedure Paste(ACol, ARow: Integer; dat: TCsvSheet); // \t
    function GetRectData(ALeft,ATop,ARight,ABottom: Integer): TCsvSheet;
    procedure UniqueKey(Index: Integer);
    // t@C
    procedure LoadFromFile(FileName: string; ProgressFunc: TCsvProgressFunc = nil; Splitter: Char = ',');
    procedure SaveToFile(FileName: string; ProgressFunc: TCsvProgressFunc = nil; Splitter: Char = ','; RetCode: string = #13#10);
    // @\
    procedure ReverseRow; // t܂ɂ
    procedure SortStr(Index: Integer);
    procedure SortNum(Index: Integer);
    procedure SortDate(Index: Integer);
    procedure RowColReverse;
    function FindStr(Index: Integer; key: string; FromIndex: Integer = 0): Integer; // ʂɊSv
    function FindWildMatch(Index: Integer; key: string; FromIndex: Integer = 0): Integer;
    function KeisenText: string;
    function OutHtmlTable(attribute: string): string;
    // Rs[̑
    procedure Assign(Source: TCsvSheet);
    function AddCsvRow(sheet: TCsvSheet; Row: Integer): Integer;
    function AddRow(row: TCsvCells): Integer;
    function getLastId(Index: Integer; min:Integer): Integer;

    // vpeB
    property Count: Integer read GetCount write SetCount;
    property ColCount: Integer read GetColCount write SetColCount;
    property Cells[ACol, ARow: Integer]: string read GetCell write SetCell;
    property AsText : string read GetAsText write SetAsText;
    property AsTabText : string read GetAsTabText write SetAsTabText;
    property UseHeader: Boolean read FUseHeader write FUseHeader;
    property Rows[ARow: Integer]: TCsvCells read GetRow;
  end;

implementation

uses StrUnit, wildcard, Variants;

{ TCsvSheet }

function TCsvSheet.AddCsvRow(sheet: TCsvSheet; Row: Integer): Integer;
var
  src, des: TCsvCells;
begin
  if sheet.list.Count <= Row then
  begin
    Result := -1;
    Exit;
  end;
  src := sheet.list.Items[Row];
  des := TCsvCells.Create ;
  if src <> nil then des.Assign(src);
  Result := list.Add(src);
end;

function TCsvSheet.AddRow(row: TCsvCells): Integer;
begin
  Result := list.Add(row);
end;

procedure TCsvSheet.Assign(Source: TCsvSheet);
var
  i: Integer;
  c, cDest: TCsvCells;
begin
  list.Count := Source.Count ;
  for i := 0 to list.Count - 1 do
  begin
    cDest := Source.list.Items[i];
    //
    c := TCsvCells.Create ;
    if cDest <> nil then c.Assign(cDest);
    list.Items[i] := c;
  end;
end;

procedure TCsvSheet.Clear;
var
  i: Integer; c: TCsvCells;
begin
  for i := 0 to list.Count - 1 do
  begin
    c := list.Items[i];
    FreeAndNil(c);
  end;
  list.Clear ;
end;

constructor TCsvSheet.Create;
begin
  list := TList.Create ;
end;

procedure TCsvSheet.DeleteCol(Index: Integer);
var
  i: Integer;
  c: TCsvCells;
begin
  for i := 0 to list.Count - 1 do
  begin
    c := list.Items[i];
    if c <> nil then
    begin
      if c.Count > Index then c.Delete(Index);
    end;
  end;
end;

procedure TCsvSheet.DeleteRow(Index: Integer);
var
  c: TCsvCells;
begin
  c := TCsvCells(list.Items[Index]);
  FreeAndNil(c);
  list.Delete(Index); 
end;

destructor TCsvSheet.Destroy;
begin
  list.Free ;
  inherited;
end;

function TCsvSheet.FindStr(Index: Integer; key: string; FromIndex: Integer): Integer;
var
  i: Integer;
  p: TCsvCells ;
begin
  Result := -1;
  for i := FromIndex to list.Count - 1 do
  begin
    p := list.Items[i];
    if p.GetValue(Index) = key then
    begin
      Result := i;
      Break;
    end;
  end;
end;

function TCsvSheet.FindWildMatch(Index: Integer; key: string;
  FromIndex: Integer): Integer;
var
  i: Integer;
  p: TCsvCells ;
  s: string;
begin
  Result := -1;
  for i := FromIndex to list.Count - 1 do
  begin
    p := list.Items[i];
    s := p.GetValue(Index);
    if WildMatchFilename(s, key) then
    begin
      Result := i;
      Break;
    end;
  end;
end;

function TCsvSheet.GetAsTabText: string;
var
  c: TCsvCells;
  row: Integer;
begin
  Result := '';
  for row := 0 to list.Count - 1 do
  begin
    c := list.Items[row];
    Result := Result + c.GetTabText + #13#10;
  end;
end;

function TCsvSheet.GetAsText: string;
var
  c: TCsvCells;
  row: Integer;
begin
  Result := '';
  for row := 0 to list.Count - 1 do
  begin
    c := list.Items[row];
    if c <> nil then
      Result := Result + c.GetCommaText + #13#10
    else
      Result := Result + #13#10;
  end;
end;

function TCsvSheet.GetCell(ACol, ARow: Integer): string;
var
  c: TCsvCells ;
begin
  if (ARow < list.Count)and(ARow >= 0) then
  begin
    c := list.Items[ARow];
    if c <> nil then
      Result := c.GetValue(ACol)
    else
      Result := ''
    ;
  end else
  begin
    Result := '';
  end;
end;

function TCsvSheet.GetColCount: Integer;
var
  c: TCsvCells; i, max: Integer;
begin
  max := 0;
  for i := 0 to list.Count - 1 do
  begin
    c := list.Items[i];
    if c <> nil then
    begin
      if max < c.Count then max := c.Count ;
    end;
  end;
  Result := max;
end;

function TCsvSheet.GetCount: Integer;
begin
  Result := list.Count ;
end;

function TCsvSheet.getLastId(Index: Integer; min:Integer): Integer;
var
  i, cid: Integer;
begin
  Result := min;
  for i := 0 to Count - 1 do
  begin
    cid := StrToIntDef(Cells[Index, i], min);
    if Result <= cid then
    begin
      Result := cid + 1;
    end;
  end;
end;

function TCsvSheet.GetRectData(ALeft, ATop, ARight,
  ABottom: Integer): TCsvSheet;
var
  x, y, cols, rows : Integer;
begin
  rows := ABottom - ATop + 1;
  cols := ARight - ALeft + 1;

  Result := TCsvSheet.Create ;
  Result.Count := rows;

  for y := 0 to rows - 1 do
  begin
    for x := 0 to cols - 1 do
    begin
      Result.Cells[x,y] := Self.Cells[ALeft+x, ATop+y];
    end;
  end;
end;

function TCsvSheet.GetRow(ARow: Integer): TCsvCells;
begin
  Result := list.Items[ARow];
end;

procedure TCsvSheet.InsertCol(Index: Integer);
var
  i: Integer;
  p: TCsvCells;
begin
  for i := 0 to Count - 1 do
  begin
    p := list.Items[i];
    if p <> nil then p.Insert(Index, '');
  end;
end;

procedure TCsvSheet.InsertRow(Index: Integer);
begin
  if Index >= list.Count then
    list.Add(TCsvCells.Create)
  else
    list.Insert(Index, TCsvCells.Create);
end;

function TCsvSheet.KeisenText: string;
const
  KEI_TOP: array[0..2]of string = ('','','');
  KEI_MID: array[0..2]of string = ('','','');
  KEI_BOM: array[0..2]of string = ('','','');
  KEIW = '';
  KEIH = '';

var
  col, row, cols, w, maxw: Integer;
  colWidth: array of Integer;

  function StrRepeat(s: string; Count: Integer): string;
  var i: Integer;
  begin
    Result := '';
    for i := 0 to Count - 1 do
      Result := Result + s;
  end;

  procedure subKeisen(L,M,R: string);
  var col: Integer;
  begin
    Result := Result + L;
    for col := 0 to cols -1 do
    begin
      Result := Result + StrRepeat(KEIW, ColWidth[col] div 2);
      if col <> (cols-1) then
        Result := Result + M
      else
        Result := Result + R + #13#10;
    end;
  end;

begin
  Result := '';
  cols := ColCount;
  if cols = 0 then Exit;
  
  SetLength(colWidth, cols);
  // JEg
  for col := 0 to cols - 1 do
  begin
    maxw := 0;
    for row := 0 to Count - 1 do
    begin
      w := Length(Cells[col,row]);
      if (w mod 2) = 1 then w := w + 1; // rQoCgȂ̂ő
      if w > maxw then maxw := w;
    end;
    colWidth[col] := maxw;
  end;

  // ܂ŏ㕔
  subKeisen(KEI_TOP[0], KEI_TOP[1], KEI_TOP[2]);

  // Z̃f[^
  for row := 0 to Count - 1 do
  begin
    // f[^
    Result := Result + KEIH;
    for col := 0 to cols - 1 do
    begin
      w := Length(Cells[col,row]);
      Result := Result + Cells[col,row];
      Result := Result + StrRepeat(' ', (ColWidth[col]-w));
      Result := Result + KEIH;
    end;
    Result := Result + #13#10;
    // 
    if row <> (Count - 1) then subKeisen(KEI_MID[0], KEI_MID[1], KEI_MID[2]);
  end;
  // ŉ[
  subKeisen(KEI_BOM[0], KEI_BOM[1], KEI_BOM[2]);
end;

procedure TCsvSheet.LoadFromFile(FileName: string;
  ProgressFunc: TCsvProgressFunc; Splitter: Char);
var
  i: Integer; c: TCsvCells; Cancel: Boolean;
  TotalSize: Integer;
  ReadBytes: Integer;
  fs: TFileStream;

  FlagEmpty: Boolean;
  buf: string;
  pbuf: PChar;

  function ReadCharFromStream: string;

    procedure getBuf;
    begin
      // obt@O
      if (pbuf = nil)or(pbuf^ = #0) then
      begin
        if fs.Position < fs.Size then
        begin
          // ܂[Ŗ߂
          ZeroMemory(@buf[1], Length(buf));
          fs.Read(buf[1], Length(buf)-1);
          pbuf := PChar(buf);
        end;
      end;
    end;

  begin
    Result := '';
    getBuf;
    if (pbuf = nil)or(pbuf^ = #0) then
    begin
      FlagEmpty := True;
      Exit;
    end;

    if pbuf^ in LeadBytes then
    begin
      Result := pbuf^;
      Inc(pbuf);
      if pbuf^ = #0 then getBuf;
      if pbuf^ = #0 then Exit;
      Result := Result + pbuf^;
      Inc(pbuf);
    end else
    begin
      // CR/LF/CRLF => CR
      if pbuf^ = #13 then
      begin
        Result := #13;
        Inc(pbuf);
        // check next
        if pbuf^ = #0 then getBuf;
        if pbuf^ = #10 then
        begin
          Inc(pbuf);
        end;
      end else
      if pbuf^ = #10 then
      begin
        Result := #13;
        Inc(pbuf);
      end else
      begin
        Result := pbuf^;
        Inc(pbuf);
      end;
    end;
  end;

  function ReadLineFromStream: string;
  var ch: string;
  begin
    Result := '';
    while not FlagEmpty do
    begin
      ch := ReadCharFromStream;
      if ch = #13 then
      begin
        Break;
      end else
      begin
        Result := Result + ch;
      end;
    end;
  end;


  procedure ReadLine;
  var
    flagStr: Boolean; p: PChar; line, cell: string;
  label
    lblRead;
  begin
    flagStr := False;
    if FlagEmpty then Exit;

    line := ReadLineFromStream;

    ReadBytes := ReadBytes + Length(line) + 2;
    p := PChar(line);
    cell := '';

    lblRead:
    while p^ <> #0 do
    begin
      if p^ in LeadBytes then
      begin
        cell := cell + p^ + (p+1)^;
        Inc(p, 2);
        Continue;
      end;
      if (p^ = '"')and((p+1)^ = '"')and(flagStr) then
      begin
        cell := cell + '"';
        Inc(p, 2);
        Continue;
      end;
      if p^ = '"' then
      begin
        Inc(p);
        flagStr := not flagStr;
        Continue;
      end;
      if (p^ = Splitter)and(flagStr = False) then
      begin
        Inc(p);
        c.Add(cell);
        cell := '';
        while (p^ = ' ') do Inc(p); // Xy[X͏ȗ
      end else
      begin
        cell := cell + p^; Inc(p);
      end;
    end;
    //
    if flagStr then
    begin
      line := ReadLineFromStream;
      if not FlagEmpty then
      begin
        ReadBytes := ReadBytes + Length(line) + 2;
        p := PChar(line);
        cell := cell + #13#10;
        goto lblRead;
      end;
    end;
    c.Add(cell);
  end;

begin
  Clear ;
  // buffer
  SetLength(buf, 65536); pbuf := nil;
  FlagEmpty := False;

  fs := TFileStream.Create(FileName, fmOpenRead);
  try
    fs.Position := 0;
    TotalSize := fs.Size;
    ReadBytes := 1;
    i := 0;
    while not FlagEmpty do
    begin
      // Progress function
      if Assigned(ProgressFunc) then
      begin
        if (i mod 100) = 0 then
        begin
          Cancel := False;
          ProgressFunc(ReadBytes, TotalSize, Cancel);
          if Cancel then Break;
        end;
      end;

      //========================================================================
      // Read File
      c := TCsvCells.Create ;
      ReadLine;
      list.Add(c);
      i := i + 1;
    end;
    if Assigned(ProgressFunc) then ProgressFunc(TotalSize, TotalSize, Cancel);
  finally
    FreeAndNil(fs);
  end;
end;

procedure TCsvSheet.MoveCol(FromI, ToI: Integer);
var
  i: Integer;
  c: TCsvCells;
begin
  for i := 0 to Count - 1 do
  begin
    c := list.Items[i];
    if c = nil then Continue;
    c.Move(FromI, ToI); 
  end;
end;

procedure TCsvSheet.MoveRow(FromI, ToI: Integer);
begin
  if list.Count <= FromI then
  begin
    list.Count := FromI + 1;
  end;
  list.Move(FromI, ToI);
end;

function TCsvSheet.OutHtmlTable(attribute: string): string;
var
  x, y: Integer;
  cols: Integer;
begin
  cols := ColCount ;
  if attribute <> '' then
  begin
    Result := '<TABLE ' + trim(attribute) + '>'#13#10;
  end else
  begin
    Result := '<TABLE>'#13#10;
  end;
  //
  for y := 0 to Count - 1 do
  begin
    Result := Result + '  <TR>'#13#10;
    for x := 0 to cols - 1 do
    begin
      Result := Result + '    <TD>'+GetCell(x,y)+'</TD>'#13#10;
    end;
    Result := Result + '  </TR>'#13#10;
  end;
  Result := Result + '</TABLE>'#13#10;
end;

procedure TCsvSheet.Paste(ACol, ARow: Integer; dat: TCsvSheet);
var
  x, y, cols, rows: Integer;
begin
  rows := dat.Count ;
  cols := dat.ColCount ;

  for y := 0 to rows - 1 do
  begin
    for x := 0 to cols - 1 do
    begin
      Cells[ACol + x, ARow + y] := dat.Cells[x, y];
    end;
  end;
end;

procedure TCsvSheet.ReverseRow;
var
  i, j: Integer;
  rows, r2: Integer;
  p: Pointer;
begin
  rows := list.Count;
  // ㉺t܂ɂ鏈
  if UseHeader then
  begin
    // 012345678 : 9
    // *i      j
    r2   := (rows-1) div 2 + 1;
    for i := 1 to r2 - 1 do
    begin
      j := rows - i;  // ЂԂΏ
      p := list.Items[i];
      list.Items[i] := list.Items[j];
      list.Items[j] := p;
    end;
  end else
  begin
    r2   := rows div 2 + 1;
    for i := 0 to r2 - 1 do
    begin
      j := rows - 1 - i;
      p := list.Items[i];
      list.Items[i] := list.Items[j];
      list.Items[j] := p;
    end;
  end;
end;

procedure TCsvSheet.RowColReverse;
var
  row, col: Integer;
  c: TCsvSheet;
begin
  c := TCsvSheet.Create;
  try
    c.Assign(Self);
    for row := 0 to Self.Count - 1 do
    begin
      for col := 0 to Self.ColCount - 1 do
      begin
        c.Cells[col, row] := Self.Cells[row, col];
      end;
    end;
    Self.Assign(c);
  finally
    c.Free;
  end;
end;

procedure TCsvSheet.SaveToFile(FileName: string;
  ProgressFunc: TCsvProgressFunc; Splitter: Char; RetCode: string);
var
  i: Integer; c: TCsvCells; Cancel: Boolean;
  s: string;
  colCount: Integer;
  fs: TFileStream;
begin
  fs := TFileStream.Create(FileName, fmCreate);
  try
    fs.Position := 0;
    colCount := GetColCount ;
    for i := 0 to list.Count - 1 do
    begin
      c := list.Items[i];
      // Progress function
      if Assigned(ProgressFunc) then
      begin
        if (i mod 30) = 0 then
        begin
          Cancel := False;
          ProgressFunc(i, list.Count-1, Cancel);
          if Cancel then Break;
        end;
      end;
      if c = nil then
        fs.Write(RetCode[1], Length(RetCode))
      else begin
        s := c.GetCommaTextEx(Splitter,colCount) + RetCode;
        fs.Write(s[1], Length(s));
      end;
    end;
    if Assigned(ProgressFunc) then ProgressFunc(100, 100, Cancel);
  finally
    FreeAndNil(fs);
  end;
end;

procedure TCsvSheet.SetAsTabText(const Value: string);
begin
  SetAsTextCustom(#9, Value);
end;

procedure TCsvSheet.SetAsText(const Value: string);
begin
  SetAsTextCustom(',', Value);
end;

procedure TCsvSheet.SetAsTextCustom(const Delimiter: Char;
  const Value: string);
var
  p: PChar;
  flagStr: Boolean;
  cell: string;
  c: TCsvCells;
begin
  Clear;
  flagStr := False;
  c := TCsvCells.Create ;
  p := PChar(Value);
  while p^ <> #0 do
  begin
    if p^ in LeadBytes then
    begin
      cell := cell + p^ + (p+1)^;
      Inc(p,2);
      Continue;
    end;

    if (flagStr) and (p^ = '"') and ((p+1)^ = '"') then
    begin
      cell := cell + '"';
      Inc(p, 2);
      Continue;
    end;

    if p^ = '"' then
    begin
      flagStr := not flagStr;
      Inc(p);
      Continue;
    end;

    if p^ = Delimiter then
    begin
      if flagStr then
      begin
        cell := cell + p^;
        Inc(p);
      end else
      begin
        c.Add(cell);
        cell := '';
        Inc(p);
      end;
      Continue;
    end;

    if (p^ in [#13, #10])and(flagStr = False) then
    begin
      if cell <> '' then
      begin
        c.Add(cell); cell := '';
      end;
      if p^ = #13 then Inc(p);
      if p^ = #10 then Inc(p);
      list.Add(c);
      c := TCsvCells.Create ;
    end else
    begin
      cell := cell + p^;
      Inc(p);
    end;
  end;
  if cell <> '' then c.Add(cell);
  if c.Count > 0 then list.Add(c) else c.Free ;
end;

procedure TCsvSheet.SetCell(ACol, ARow: Integer; const Value: string);
begin
  if ARow >= list.Count then
  begin
    list.Count := ARow + 1;
    list.Items[ARow] := TCsvCells.Create;
  end;
  if list.Items[ARow] = nil then list.Items[ARow] := TCsvCells.Create ;
  TCsvCells(list.Items[ARow]).Values[ACol] := Value;
end;

procedure TCsvSheet.SetColCount(const Value: Integer);
var
  p: TCsvCells;
begin
  if list.Count = 0 then list.Count := 1;
  p := list.Items[0];
  if p = nil then p := TCsvCells.Create ;
  p.Count := Value;
  list.Items[0] := p;
end;

procedure TCsvSheet.SetCount(const Value: Integer);
begin
  list.Count := Value;
end;

var
  CustomSortIndex: Integer;

function CustomSortString(Item1, Item2: Pointer): Integer;
var
  c1, c2: TCsvCells;
begin
  c1 := TCsvCells(Item1);
  c2 := TCsvCells(Item2);
  Result := CompareText(
    c1.Values[CustomSortIndex],
    c2.Values[CustomSortIndex]);
end;

function CustomSortNumber(Item1, Item2: Pointer): Integer;
var
  c1, c2: TCsvCells;
  n1, n2: Extended;
begin
  c1 := TCsvCells(Item1);
  c2 := TCsvCells(Item2);
  n1 := StrToValue(c1.Values[CustomSortIndex]);
  n2 := StrToValue(c2.Values[CustomSortIndex]);
  Result := Trunc(n1 - n2);
end;

function CustomSortDate(Item1, Item2: Pointer): Integer;
var
  c1, c2: TCsvCells;
  d1, d2: TDateTime;
begin
  c1 := TCsvCells(Item1);
  c2 := TCsvCells(Item2);
  try
    d1 := VarToDateTime(c1.Values[CustomSortIndex]);
  except
    d1 := 0;
  end;
  try
    d2 := VarToDateTime(c2.Values[CustomSortIndex]);
  except
    d2 := 0;
  end;
  // tɕϊłȂ΁A0 Ƃ
  if d1 = d2 then Result := 0 else
  if d1 > d2 then
  begin
    Result := 1;
  end else
  begin
    Result := -1;
  end;
end;

procedure TCsvSheet.SortDate(Index: Integer);
var
  p: Pointer;
begin
  CustomSortIndex := Index;
  if list.Count = 0 then Exit;

  if UseHeader then
  begin
    p := list.Items[0];
    list.Delete(0);
    list.Sort(@CustomSortDate);
    list.Insert(0, p);
  end else
  begin
    list.Sort(@CustomSortDate);
  end;
end;

procedure TCsvSheet.SortNum(Index: Integer);
var
  p: Pointer;
begin
  CustomSortIndex := Index;
  if list.Count = 0 then Exit;

  if UseHeader then
  begin
    p := list.Items[0];
    list.Delete(0);
    list.Sort(@CustomSortNumber);
    list.Insert(0, p);
  end else
  begin
    list.Sort(@CustomSortNumber);
  end;
end;

procedure TCsvSheet.SortStr(Index: Integer);
var
  p: Pointer;
begin
  CustomSortIndex := Index;
  if list.Count = 0 then Exit;

  if UseHeader then
  begin
    p := list.Items[0];
    list.Delete(0);
    list.Sort(@CustomSortString);
    list.Insert(0, p); 
  end else
  begin
    list.Sort(@CustomSortString);
  end;
end;

procedure TCsvSheet.TrimBottom;
var
  c: TCsvCells;
begin
  while list.Count > 0 do
  begin
    c := list.Items[ list.Count - 1 ];

    if (c = nil) or (c.AllBlank) then
    begin
      list.Delete(list.Count - 1);
      FreeAndNil(c);
    end else
    begin
      Break;
    end;
  end;
end;

procedure TCsvSheet.TrimRight;
var
  i: Integer;
  c: TCsvCells ;
begin
  for i := 0 to list.Count - 1 do
  begin
    c := list.Items[i];
    if c=nil then Continue;
    c.TrimLast ;
  end;
end;

procedure TCsvSheet.UniqueKey(Index: Integer);
var
  i, j: Integer;
begin
  i := 0;
  while i < Count do
  begin
    j := i + 1;
    while j < Count do
    begin
      if GetCell(Index, i) = GetCell(Index, j) then
      begin
        DeleteRow(j);
      end else
      begin
        Inc(j);
      end;
    end;
    Inc(i);
  end;
end;

{ TCsvCells }

function TCsvCells.Add(Value: string): Integer;
var
  p: PChar;
begin
  GetMem(p, Length(Value) + 1);
  StrCopy(p, PChar(Value)); 
  Result := list.Add(p);
end;

function TCsvCells.AllBlank: Boolean;
var
  i: Integer;
  s: string;
begin
  Result := False;
  for i := 0 to list.Count - 1 do
  begin
    s := GetValue(i);
    if s <> '' then Exit; //***
  end;
  Result := True;
end;

procedure TCsvCells.Assign(Source: TCsvCells);
var i: Integer; pSrc, pDes: PChar;
begin
  list.Count := Source.Count ;
  for i := 0 to Source.Count - 1 do
  begin
    pSrc := Source.list.Items[i];
    if pSrc <> nil then
    begin
      GetMem(pDes, StrLen(pSrc)+1);
      StrCopy(pDes, pSrc);
    end else
      pDes := nil;
    //
    list.Items[i] := pDes;
  end;
end;

procedure TCsvCells.Clear;
var i: Integer; p: PChar;
begin
  for i := 0 to list.Count - 1 do
  begin
    p := list.Items[i];
    if p <> nil then Dispose(p);
  end;
  list.Clear ;
end;

constructor TCsvCells.Create;
begin
  list := TList.Create ;
end;

procedure TCsvCells.Delete(Index: Integer);
var
  p: PChar;
begin
  p := list.Items[Index];
  if p<>nil then Dispose(p);
  list.Delete(Index); 
end;

destructor TCsvCells.Destroy;
begin
  Clear;
  list.Free ;
  inherited;
end;

function TCsvCells.GetCommaText: string;
var
  i: Integer;

  function chkValue(s: string): string;
  begin
    Result := s;
    if (JPosM('"', Result) > 0)or(JPosM(#13, Result) > 0)or(JPosM(#10, Result) > 0)then
    begin
      Result := '"' + JReplace(Result, '"', '""', True) + '"';
    end else
    if (JPosM(',', Result) > 0)or(JPosM(' ', Result) > 0) then
    begin
      Result := '"' + Result + '"';
    end else
    if (Result = 'ID')or(Result = 'id') then
    begin
      Result := '"' + Result + '"';
    end;
  end;

begin
  Result := '';
  for i := 0 to list.Count - 1 do
  begin
    Result := Result + chkValue(GetValue(i)) + ',';
  end;
  if Result <> '' then System.Delete(Result, Length(Result), 1);
end;

function TCsvCells.GetCommaTextEx(Splitter: Char; ColCount: Integer): string;
var
  i: Integer;

  function chkValue(s: string): string;
  begin
    Result := s;
    if (JPosM('"', Result) > 0)or(JPosM(#13, Result) > 0)or(JPosM(#10, Result) > 0)then
    begin
      Result := '"' + JReplace(Result, '"', '""', True) + '"';
    end else
    if (JPosM(Splitter, Result) > 0)or(JPosM(' ', Result) > 0) then
    begin
      Result := '"' + Result + '"';
    end else
    if (Result = 'ID')or(Result = 'id') then
    begin
      Result := '"' + Result + '"';
    end;
  end;

begin
  Result := '';
  if ColCount < 0 then ColCount := list.Count ;
  for i := 0 to ColCount - 1 do
  begin
    Result := Result + chkValue(GetValue(i)) + Splitter;
  end;
  if Result <> '' then System.Delete(Result, Length(Result), 1);
end;

function TCsvCells.GetCount: Integer;
begin
  Result := list.Count ;
end;


function TCsvCells.GetTabText: string;
var
  i: Integer;

  function chk(s: string): string;
  begin
    if (JPosM('"', s) > 0)or(JPosM(#13, s) > 0)or(JPosM(#10, s) > 0) then
    begin
      Result := '"' + JReplace(s, '"', '""', True) + '"';
    end else
    if JPosM(#9, s) > 0 then
    begin
      Result := '"' + s + '"';
    end else
    begin
      Result := s;
    end;
  end;

begin
  Result := '';
  for i := 0 to list.Count - 1 do
  begin
    Result := Result + chk(PChar(list.Items[i])) + #9;
  end;
  System.Delete(Result,Length(Result),1);
end;

function TCsvCells.GetValue(Index: Integer): string;
begin
  if (Index < list.Count)and(Index >= 0) then
  begin
    Result := string(PChar( list.Items[Index] ));
  end else
  begin
    Result := '';
  end;
end;

procedure TCsvCells.Insert(Index: Integer; Value: string);
var
  p: PChar;
begin
  if Value <> '' then
  begin
    GetMem(p, Length(Value)+1);
    StrCopy(p, PChar(Value));
  end else
    p := nil;
  if Index >= Count then list.Count := Index;
  list.Insert(Index, p);
end;

procedure TCsvCells.Move(FromI, ToI: Integer);
begin
  if FromI >= list.Count then
  begin
    list.Count := FromI + 1;
  end;
  list.Move(FromI, ToI);
end;

procedure TCsvCells.SetCount(const Value: Integer);
begin
  list.Count := Value;
end;

procedure TCsvCells.SetValue(Index: Integer; const Value: string);
var p: PChar;
begin
  if Index >= list.Count then
  begin
    list.Count := Index + 1; // Lт
  end;
  GetMem(p, Length(Value) + 1);
  StrCopy(p, PChar(Value));
  list.Items[Index] := p;
end;

procedure TCsvCells.TrimLast;
var
  i: Integer;
begin
  i := list.Count - 1;
  while i >= 0 do
  begin
    if Trim(GetValue(i)) = '' then
    begin
      Delete(i);
      Dec(i);
    end else
    begin
      Break;
    end;
  end;
end;

end.
